// CRayTraceCamera.cpp: the implementation of the CRayTraceCamera class.
// A class that makes ray-traced images of a scene.
// Written for COMPSCI 372, May 2003, Richard Lobb.
//////////////////////////////////////////////////////////////////////

#include <windows.h>
#include <gl/gl.h>
#include <gl/glu.h>
#include <gl/glut.h>
#include "Geometry.h"
#include "CRayTraceCamera.h"
#include "SceneDescription.h"
#include <math.h>
#include <iostream>

#define SMALL_IMAGE_DEBUG 0  // A few debugging controls  ...
#define DEBUG_RAY_ROW 100     // ... so we can trace the ...
#define DEBUG_RAY_COL 150    // ... execution of a small ...
#define SMALL_IMAGE_SIZE 5   // .. subimage.

CRayTraceCamera::CRayTraceCamera(const CVec3df& eyePosition,
            const CVec3df& lookat, const float fovDegs) {
    eyePos = eyePosition;
    fov = fovDegs;
    n = (eyePosition - lookat).normalise();  // The usual UVN (view space) "n" vector
    CVec3df up(0,1,0);						 // The "UP" vector
    u = cross(up,n).normalise();			 // The UVN u vector
    v = cross(n,u);							 // The UVN v vector
}

void CRayTraceCamera::rayTrace(CScene& scene) {
    clearStats();
    float H = (float) (tan(fov/2 * PI / 180));  // Half height (in world coords)
    float nCols = glutGet(GLUT_WINDOW_WIDTH);   // Number of pixel columns
    float nRows = glutGet(GLUT_WINDOW_HEIGHT);  // Number of pixel rows
    float W = H * nCols / nRows;                // Half width (world coords)
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0, nCols, 0, nRows);
    glDisable(GL_LIGHTING);
    int rowMin = 0, rowMax = (int) nRows, colMin = 0, colMax = (int) nCols;
    if (SMALL_IMAGE_DEBUG) {  // Debug capability -- trace a small sub-image
        rowMin = DEBUG_RAY_ROW;
        rowMax = DEBUG_RAY_ROW + SMALL_IMAGE_SIZE;
        colMin = DEBUG_RAY_COL;
        colMax = DEBUG_RAY_COL + SMALL_IMAGE_SIZE;
    }
    for (int row = rowMin; row < rowMax ; row++) {
        glBegin(GL_POINTS);  // Do each scan line as an OpenGL "POINTS" primitive
        for (int col = colMin; col < colMax; col++) {
			// The next few statements compute the ray direction vector by summing its
			// UVN coordinates. "-n" takes you to the viewplane and the "u" and
			// "v" terms select the required pixel in the viewplane.
			float uCoord = (2 * col/nCols - 1) * W;	// Horizontal distance of pixel from centre of viewplane
			float vCoord = (2 * row/nRows - 1) * H; // Vertical distance of pixel from centre of viewplane
            CVec3df rayDir = (-n + u * uCoord + v * vCoord).normalise();
            CRay ray(eyePos, rayDir);
            if (SMALL_IMAGE_DEBUG) cout << "Tracing pixel (" << col << "," << row << "), ray = " << ray << endl;
            CColour pixelColour = scene.shade(ray);
            glColor3fv(pixelColour.getGLColour());
            glVertex2f(col, row);
        }
        glEnd();
     }
     cout << "\n\nRendering complete\n";
     cout << "NumShadeRays = " << CScene::numShadeRays <<
          ", numShadowTestRays = " << CScene::numShadowTestRays << ", numIntersectionTests = "
          << CSceneObject::numHitCalls << ", numLeafsVisited = " << CBSPLeaf::numLeafsVisited << endl;
}

void CRayTraceCamera::clearStats() {
	CBSPLeaf::numLeafsVisited = 0;
    CSceneObject::numHitCalls = 0;
    CScene::numShadeRays = 0;
    CScene::numShadowTestRays = 0;
}


