// Transformation.h -- the header file for the affine transformation class.

// Based on Burkhard's CMatrix3df class, but is a 4 x 4 matrix in order
// to handle translations.

// Written for COMPSCI 372, Richard Lobb, May 2003.

#include <iostream>
#include <math.h>
#include "Geometry.h"
#include "Utilities.h"
#include "CTransformation.h"

CTransformation::CTransformation() { // Default constructor == identity
    for(int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++)
          m[i][j] = i == j ? 1.0f : 0;
}

CTransformation::CTransformation(float m2[4][4]) {
    for(int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++)
          m[i][j] = m2[i][j];
}

CTransformation::CTransformation(const CTransformation& m2) {
    for(int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++)
          m[i][j] = m2.m[i][j];
}

CTransformation& CTransformation::operator=(const CTransformation& m2) {
// Assignment operator
    for(int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++)
          m[i][j] = m2.m[i][j];
    return *this;
}

CTransformation& CTransformation::operator*=(const CTransformation& m2) {
    float prod[4][4];
    for(int i = 0; i < 4; i++){
       for (int j = 0; j < 4; j++) {
          prod[i][j] = 0;
          for (int k = 0; k < 4; k++)
             prod[i][j] += m[i][k] * m2.m[k][j];
       }
    }
    for(int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++)
          m[i][j] = prod[i][j];
    return *this;
}

CTransformation& CTransformation::operator*=(float scalar) {
    for(int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++)
          m[i][j] *= scalar;
    return *this;
}

CTransformation& CTransformation::operator/=(float scalar)  {
    for(int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++)
          m[i][j] /= scalar;
    return *this;
}

CTransformation operator*(float scalar, const CTransformation& m1) {
    CTransformation result(m1);
    result *= scalar;
    return result;
}

CTransformation operator*(const CTransformation& m1, float scalar) {
    CTransformation result(m1);
    result *= scalar;
    return result;
}

CTransformation operator/(const CTransformation& m1, float scalar) {
    CTransformation result(m1);
    result /= scalar;
    return result;
}

CVec3df operator*(const CTransformation& m1, const CVec3df& v1) {
    float s[] = {0,0,0,0};
    for (int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++)
          s[i] += m1.m[i][j] * (j == 3 ? 1 : v1[j]);
    return CVec3df(s[0]/s[3], s[1]/s[3], s[2]/s[3]);
}

CVec3df operator*(const CVec3df& v1, const CTransformation& m1) {
    float s[] = {0,0,0,0};
    for (int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++)
          s[i] += m1.m[i][j] * (j == 3 ? 1 : v1[j]);
    return CVec3df(s[0]/s[3], s[1]/s[3], s[2]/s[3]);
}

CTransformation operator*(const CTransformation& m1, const CTransformation& m2) {
    CTransformation result(m1);
    result *= m2;
    return result;
}

// Transformation of a VECTOR!! This differs from the usual "*"
// operator, which is used to transform POINTS (albeit represented
// by position vectors of type CVec3df). This method ignores the
// the last row and column of the 4 x 4 matrix.
CVec3df CTransformation::transformVector(CVec3df& v) {
    float s[] = {0,0,0};
    for (int i = 0; i < 3; i++)
       for (int j = 0; j < 3; j++)
          s[i] += m[i][j] * v[j];
    return CVec3df(s[0], s[1], s[2]);
}

CTransformation CTransformation::transpose(void) const {
    float trans[4][4];
    for(int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++) {
          trans[i][j] = m[j][i];
       }
    return CTransformation(trans);
}

// A couple of helper functions for the matrix inverse

int pivotRow(int row, float m[4][8]) {
// Returns the row number (>= row) containing the largest element in the row-th column
    float max = (float) fabs(m[row][row]);
    int result = row;
    for (int i = row + 1; i < 4; i++) {
        float pivotMag = (float) fabs(m[i][row]);
        if ( pivotMag > max) {
            max = pivotMag;
            result = i;
        }
    }
    return result;
}
    
void swapRows(int i, int j, float m[4][8] ) {
// Interchange rows i and j of the 4 x 8 inverse-in-the-making.
   if (i == j ) return;
   for (int col = 0; col < 8; col++) {
       float temp = m[i][col];
       m[i][col] = m[j][col];
       m[j][col] = temp;
   }
}

// Now the inverse, calculated by Gauss-Jordan elimination
// [Yeah, I know the numerical analysts wouldn't approve but it's easy to write!]
CTransformation CTransformation::inverse(void) const {
   float mm[4][8];   // Copy of M in left half, result lands up in right half
   for (int i = 0; i < 4; i++) {
       for (int j = 0; j < 4; j++) {
           mm[i][j] = m[i][j];
           mm[i][j+4] = i == j ? 1.0f : 0.0f; // Identity in RH half
       }
   }
   for (int row = 0; row < 4; row++) {  // For each row
       int pRow = pivotRow(row, mm);    // Find the row with the largest pivot element
       swapRows(pRow, row, mm);         // Swap that row with the current (so pivot is on diagonal)
       float pivot = mm[row][row];     // Get the pivot element
       if (fabs(pivot) < 1.e-6)
            displayMessage(ERROR_MESSAGE, "Inverting a singular matrix");
       for (int col = row; col < 8; col++)
           mm[row][col] /= pivot;       // Scale down current row so diagonal element = 1
           for (int row2 = 0; row2 < 4; row2++) {
               if (row2 != row && mm[row2][row] != 0) {
               float scale = mm[row2][row];
               for (int col2 = row; col2 < 8; col2++)
                   mm[row2][col2] -= mm[row][col2] * scale;
           }
       }
   }
   float inv[4][4];
   for (int i = 0; i < 4; i++)
       for (int j = 0; j < 4; j++)
           inv[i][j] = mm[i][j+4];
   return CTransformation(inv);
}

ostream& operator<<(ostream& s, const CTransformation& m1)
{
    s << "\n((" << m1.m[0][0] << ',' << m1.m[0][1] << ',' << m1.m[0][2] << ',' << m1.m[0][3] << ")\n";
    s << " (" << m1.m[1][0] << ',' << m1.m[1][1] << ',' << m1.m[1][2] << ',' << m1.m[1][3] << ")\n";
    s << " (" << m1.m[2][0] << ',' << m1.m[2][1] << ',' << m1.m[2][2] << ',' << m1.m[2][3] << ")\n";
    s << " (" << m1.m[3][0] << ',' << m1.m[3][1] << ',' << m1.m[3][2] << ',' << m1.m[3][3] << "))";
    return s;
}

istream& operator>>(istream& s, CTransformation& m1)
{
    char c = 0;

    s >> c;
    if (c != '(')
        displayMessage(ERROR_MESSAGE, "CTransformation::>>. bad input format");
    for(int i=0;i<4;i++)  {
        s >> c;
        if (c != '(')
            displayMessage(ERROR_MESSAGE, "CTransformation::>>. bad input format");
        for(int j=0;j<4;j++) {
            s >> m1.m[i][j];
            if (j!=3) {
                s >> c;
                if (c != ',')
                    displayMessage(ERROR_MESSAGE, "CTransformation::>>. bad input format");
            }
        }
        s >> c;
        if (c != ')')
            displayMessage(ERROR_MESSAGE, "CVec3df::>>. bad input format");
        if (i!=3) {
            s >> c;
            if (c != ',')
                displayMessage(ERROR_MESSAGE, "CTransformation::>>. bad input format");
        }
    }
    s >> c;
    if (c != ')')
        displayMessage(ERROR_MESSAGE, "CTransformation::>>. bad input format");
    return s;
};

// Now those subclasses.

// Now some subclasses, defined in order to provide nicely named constructors
// for various common transformations.

CTranslate::CTranslate(CVec3df& shift) {
  float tx[] = {shift[X], shift[Y], shift[Z], 1};
  for (int i = 0; i < 4; i++)
      for (int j = 0; j < 4; j++)
         if (j == 3)
            m[i][j] = tx[i];
         else
            m[i][j] = i == j ? 1.0f : 0.0f;
};

CTranslate:: CTranslate(float dx, float dy, float dz) {
  float tx[] = {dx, dy, dz, 1};
  for (int i = 0; i < 4; i++)
      for (int j = 0; j < 4; j++)
         if (j == 3)
            m[i][j] = tx[i];
         else
            m[i][j] = i == j ? 1.0f : 0.0f;
};

CScale::CScale(float xScale, float yScale, float zScale) {
   for (int i = 0; i < 4; i++)
      for (int j = 0; j < 4; j++)
          m[i][j] = 0;
   m[0][0] = xScale;
   m[1][1] = yScale;
   m[2][2] = zScale;
   m[3][3] = 1;
};

CRotate::CRotate(int axis, float angleInDegs) { // Axis = 0, 1, 2 for x, y, z.
    float theta = angleInDegs * Pi / 180;
    float c = (float) cos(theta), s = (float) sin(theta);
    for (int i = 0; i < 4; i++)
        for (int j = 0; j < 4; j++) {
           if (i == 3 || j == 3 || i == axis || j == axis)
               m[i][j] = i == j ? 1.0f : 0.0f;  // All this stuff is the identity matrix
           else if (i == j)
              m[i][j] = c;     // Cos theta on diagonal
           else m[i][j] = i == (axis+1)%3 ? -s : s;
         }
};

CRotate::CRotate(CVec3df& axis, float angleInDegs) {
    float x = axis[X], y = axis[Y], z = axis[Z];
    float theta = Pi * angleInDegs/ 180;
    float c = (float) cos(theta);
    float s = (float) sin(theta);
    float t = 1 - c;
    m[0][0] = t*x*x+c;
    m[0][1] = t*x*y-s*z;
    m[0][2] = t*x*z+s*y;
    m[0][3] = 0;
    m[1][0] = t*x*y+s*z;
    m[1][1] = t*y*y+c;
    m[1][2] = t*y*z-s*x;
    m[1][3] = 0;
    m[2][0] = t*x*z-s*y;
    m[2][1] = t*y*z+s*x;
    m[2][2] = t*z*z+c;
    m[2][3] = 0;
    m[3][0] = 0;
    m[3][1] = 0;
    m[3][2] = 0;
    m[3][3] = 1;
};


CIdentity::CIdentity() {
   for (int i = 0; i < 4; i++)
      for (int j = 0; j < 4; j++)
         m[i][j] = i == j ? 1.0f : 0.0f;

};

