// Implementation of RayStuff -- i.e. the class CRay and
// related ray-object intersection classes.

// Richard Lobb, May 2003.

#include "RayStuff.h"
#include "geometry.h"
#include "utilities.h"

#define TOLERANCE 0.000001f

// CRay
// ====
CRay::CRay(CVec3df& startPoint, CVec3df& direction) {
    start = startPoint;
    dir = direction;
    recursionDepth = 0;
    tmin = 0;
    tmax = 1000000000;
}

CRay::CRay(CVec3df& startPoint, CVec3df& direction, int recursionDepth) {
    start = startPoint;
    dir = direction;
    this->recursionDepth = recursionDepth;
    tmin = 0;
    tmax = 1000000000;
}

CRay::CRay(CVec3df& startPoint, CVec3df& direction, int recursionDepth, float tmin, float tmax) {
    start = startPoint;
    dir = direction;
    this->recursionDepth = recursionDepth;
    this->tmin = tmin;
    this->tmax = tmax;
}    

void CRay::clipToBBox(CBBox3f& box) {
// Clip to a bounding box.
// Uses a general algorithm which could be made MUCH for efficient for an
// axis aligned box but I don't use it much so can't be bothered.
 /*   clipToHS(-1,0,0, -box.xmin); cout << (*this) << endl;
    clipToHS(1,0,0,box.xmax);cout << (*this) << endl;
    clipToHS(0,-1,0,-box.ymin);cout << (*this) << endl;
    clipToHS(0,1,0,box.ymax);cout << (*this) << endl;
    clipToHS(0,0,-1,-box.zmin);cout << (*this) << endl; 
    clipToHS(0,0,1,box.zmax);cout << (*this) << endl;
 */
    clipToHS(-1,0,0, -box.xmin);
    clipToHS(1,0,0,box.xmax);
    clipToHS(0,-1,0,-box.ymin);
    clipToHS(0,1,0,box.ymax);
    clipToHS(0,0,-1,-box.zmin);
    clipToHS(0,0,1,box.zmax);
}

void CRay::clipToHS(float nx, float ny, float nz, float d) {
// Clip this to inside of plane with normal (nx,ny,nz), distance from origin d.
    CVec3df normal(nx,ny,nz);
    float perpVelocity = dir.dot(normal);
    float distanceToPlane = d - start.dot(normal);
    if (abs(perpVelocity) < TOLERANCE) {  // Is ray parallel to plane?
        if (distanceToPlane < 0)       // Yes. Is it outside plane?
            tmin = tmax = 0;           // Yes. null it.
        return;
    }
    float tHit = distanceToPlane / perpVelocity;  // t when we hit the plane
    if (perpVelocity > 0) // If we're heading outside
       tmax = min(tmax, tHit);  // Adjust upper limit
    else
       tmin = max(tmin, tHit);  // Else adjust lower limit
} 


CVec3df CRay::p(float t) {  // p(t) on ray.
   return CVec3df(start + dir * t);
}

CRay CRay::transformed(CTransformation& T) {
   CVec3df newStart = T * start;
   CVec3df newDir = T.transformVector(dir);
   return CRay(newStart, newDir);
}

ostream& operator<<(ostream& s, const CRay& r) {
   s << "(" << r.start << "," << r.dir << "[" << r.tmin << "," << r.tmax << "])";
   return s;
}

// CHitInfo
// ========

void CHitInfo::set(float t, CSceneObject* pObj) {
   this->t = t;
   this->pObj = pObj;
}

// CIntersection
// =============
void CIntersection::set(CIntersection& other) {     // Copies all data from other to this
    numHits = other.numHits;
    for (int i = 0; i < numHits; i++)
       hit[i].set(other.hit[i]);
}

void CIntersection::addHit(float t, CSceneObject* pObj) {
    if (numHits == MAX_HITS)
       displayMessage(ERROR_MESSAGE, "Too many hits!");
    else
       hit[numHits++].set(t, pObj);
}


