//////////////////////////////////
//
// Header file for the various Ray Tracer support classes.
//
// Written for COMPSCI 372, May 2003. Richard Lobb.
//
// Based on Hill's text and source code, but simplified and altered
// to use Burkhard's Geometry classes. I altered the Scene and SceneObject
// classes quite a lot too :)

// Richard Lobb, May 2003.
//
//////////////////////////////////

#if !defined(RAY_STUFF_372FC)
#define RAY_STUFF_372FC

#include "Geometry.h"
#include "Utilities.h"
#include "CColour.h"
#include "CTransformation.h"
#include <math.h>

static const int MAX_HITS = 10;
// Maximum number of hits between a ray and a scene object. In my implementation
// the whole scene is a single scene object, so this is far too small for
// handling a large-scale scene. However, the whole ray tracer is inadequate
// for handling a large-scale scene, so there's no real problem here :)

static const int MAX_RECURSION_DEPTH = 3;
// Maximum no. of recursions on reflection and/or refractions.

// A Ray is a directed line from a given point in a given direction.
// Note that the direction vector may not be normalised, e.g. if the
// ray has been transformed.
class CRay {
  public:
     CVec3df start;
     CVec3df dir;
     int recursionDepth;
     float tmin, tmax;        // t-range for subrays.
     CRay() {};
     CRay(CVec3df& startPoint, CVec3df& direction);
     CRay(CVec3df& startPoint, CVec3df& direction, int recursionDepth);
     CRay(CVec3df& startPoint, CVec3df& direction, int recursionDepth, float tmin, float tmax);
     CVec3df p(float t);  // p(t) on ray.
     CRay transformed(CTransformation& t);   // Transformed ray
     void clipToBBox(CBBox3f& bbox);         // Clip to a bounding box 
     bool isNull() { return tmin >= tmax; }  // An empty subray
     friend ostream& operator<<(ostream& s, const CRay& r);
  protected:
     void clipToHS(float nx, float ny, float nz, float d); 
     // Clips this to the inside region of plane n = (nx,ny,nz), d.
};

class CSceneObject; // Forward reference.

// A HitInfo object defines a single ray/surface intersection.
// I've deleted most fields from Hill's version since they can be obtained
// by other methods, e.g. see CSceneObject::getHitInfo().
class CHitInfo {
  public:
	 float t;    	         // the hit time
	 CSceneObject* pObj;     // pointer to the object hit
     CHitInfo() {};          // Constructor does nothing -- use "set"
     void set(float t, CSceneObject* pObj);
     void set(CHitInfo& other) { set(other.t, other.pObj); }
};

// An Intersection object defines the intersection of a ray with a scene object
// It just contains multiple "HitInfo" objects in an array. [Probably not how
// I'd go about it myself, but I'll stick to Hill's design.]
// The hit list, "hits", should always be in numerical order by t value.
class CIntersection {
  public:
     int numHits;  // No. of defined entries in the "list"
     CHitInfo hit[MAX_HITS];  // Array of hit records
     CIntersection() { numHits = 0; }
     void clear() { numHits = 0; }       // Clear the list
     void set(CIntersection& other);     // Copies all data from other to this
     void addHit(float t, CSceneObject* obj);  // Add's a hit to the list.
     float nthT(int n) {  // t value in the nth hit, or infinity if ...
         // ... n > numHits. n is zero origin and must be positive.
         return n < numHits ? hit[n].t : (float) HUGE_VAL;
     }
};


#endif  //(RAY_STUFF_372FC)

