// Implementation of the Scene Description classes including:
//
//        CMaterial
//        CLight
//        CScene
//        CSceneObject and its subclasses
//        BSP tree stuff

// This version is the one for 715 Problem Sheet 4.

#include "SceneDescription.h"
#include <math.h>

const float TOLERANCE = 0.000002f;
const float BIG_VALUE = 1.0e30f;

// CMaterial
// =========
CMaterial::CMaterial(CColour& ambient, CColour& diffuse, CColour& specular, float shininess) {
    this->ambient = ambient;
    this->diffuse = diffuse;
    this->specular = specular;
    this->shininess = shininess;
    reflectivity = 0;
    transparency = 0;
}

CMaterial::CMaterial(CColour& ambient, CColour& diffuse, CColour& specular,
       float shininess, float reflectivity, float transparency) {
    this->ambient = ambient;
    this->diffuse = diffuse;
    this->specular = specular;
    this->shininess = shininess;
    this->reflectivity = reflectivity;
    this->transparency = transparency;
}

CMaterial::CMaterial() {  // Default constructor yields a black material.
    CColour black(0,0,0);
    ambient = diffuse = specular = black;
    shininess = 1;
    reflectivity = 0;
    transparency = 0;
}

// CSceneObject
// ============

void CSceneObject::setTransformation(CTransformation& T) {
    this->T = T;
    TInverse = T.inverse();
    TInvTranspose = TInverse.transpose();
    CVec3df cnr0 = T * bbox.corner(0);  // Compute the new bounding box
    CVec3df cnr7 = T * bbox.corner(7);  // ... by transforming diag. opp. corners
    bbox.set(cnr0, cnr7);
}

int CSceneObject::whatSideOf(CPlane3df& plane) {
    // Default implementation is to use bounding box.
    int numVerticesInside = 0, numVerticesOutside = 0;
    for (int i = 0; i < 8; i++) {
        float d = plane.distance(bbox.corner(i));
        if (d > TOLERANCE)
           numVerticesOutside++;
        else if (d < -TOLERANCE)
           numVerticesInside++;
    }
    if (numVerticesInside == 8) return -1;
    if (numVerticesOutside == 8) return 1;
    return 0;
}


// CSphere
// =======
CSphere::CSphere(CVec3df& centre, float radius, CMaterial& mat) {
   CTransformation tx = CTranslate(centre) * CScale(radius, radius, radius);
   bbox.set(-1,1,-1,1,-1,1);
   setTransformation(tx);
   this->material = mat;
}

CSphere::CSphere(CTransformation& T, CMaterial& mat) {
   bbox.set(-1,1,-1,1,-1,1);
   setTransformation(T);
   this->material = mat;
}

bool CSphere::hit(CRay& ray, CIntersection& inter) {
     // The sphere-ray intersection routine
   inter.clear();
   numHitCalls++;
   transformedRay = ray.transformed(TInverse);
     // Ray transformed back into primitive (unit sphere) space.
   float a = dot(transformedRay.dir, transformedRay.dir);
   float b = 2 * dot(transformedRay.start, transformedRay.dir);
   float c = dot(transformedRay.start, transformedRay.start) - 1;
   float discrim = b * b - 4 * a * c;
   if (discrim <= 0) return false;     // Ray misses sphere
   float t0 = (-b - (float) sqrt(discrim))/(2 * a);
   float t1 = (-b + (float) sqrt(discrim))/(2 * a);
   if (t1 <= 0) return false;          // Sphere is behind ray start
   if (t0 > 0)                         // True unless start inside sphere!
      inter.addHit(t0, this);          // Add to intersection list
   inter.addHit(t1, this);
   pLastRay = &ray;                    // Save a pointer to the ray
                                       // [See getHitInfo]
   return true;
}

void CSphere::getHitData(CRay& ray, float t, CVec3df& hitPoint, CVec3df& hitNormal) {
    if (&ray != pLastRay)
       displayMessage(ERROR_MESSAGE, "Unexpected call to getHitData");
    hitPoint = ray.p(t);
//   hitNormal = (TInvTranspose * transformedRay.p(t)).normalise(); // NO NO NO RICHARD!!!
    CVec3df genericNormal = transformedRay.p(t); // This is a vector not a point!!
    hitNormal = TInvTranspose.transformVector(genericNormal).normalise();
	// cout << "Got hit at " << hitPoint << ", with normal " << hitNormal << endl;
}


// CTaperedCylinder
// ================
CTaperedCylinder::CTaperedCylinder(float s, CTransformation& M, CMaterial& mat) {
    this->material = mat;
    this->s = s;
    float maxRadius = max(1.0f,s);
    bbox.set(-maxRadius, maxRadius, -maxRadius, maxRadius,0,1);
    setTransformation(M);
}

float min(float a, float b) { return a <= b ? a : b; }
float max(float a, float b) { return a >= b ? a : b; }

bool CTaperedCylinder::hit(CRay& r, CIntersection& inter) {
    inter.clear();
    numHitCalls++;
    transformedRay = r.transformed(TInverse);
    float sx = transformedRay.start[X], sy = transformedRay.start[Y], sz = transformedRay.start[Z];
    float cx = transformedRay.dir[X], cy = transformedRay.dir[Y], cz = transformedRay.dir[Z];
    float d = (s - 1)*cz;
    float F = 1 + (s-1)*sz;
    float a = cx*cx + cy*cy - d*d;
    float b = 2*cx*sx + 2*cy*sy - 2*F*d;
    float c = sx*sx + sy*sy - F*F;
    float discrim = b * b - 4 * a * c;
    if (discrim <= 0) return false;     // Ray misses cone part
    float t0;
    if (b > 0) t0 = (-b - (float) sqrt(discrim))/(2 * a);
    else t0 = (-b + (float) sqrt(discrim))/(2 * a);
    float t1 = c / (a * t0);
    float tMin = fmin(t0,t1);
    float tMax = fmax(t0,t1);
    if (a < 0) { // Ray inside cone at +/- infinity?
        // If yes, full line intersection is actually {[-inf, tMin],[tMax, inf]}
        // Need to get the "low-z" of those two (for the reqd bit of cone)
       if (cz * (s-1) < 0) { // sign(rayDir) != sign(s-1) ?
          tMax = tMin;
          tMin = (float) -BIG_VALUE;
        } 
       else {
          tMin = tMax;
          tMax = BIG_VALUE;
       }
    }  
    tz0 = -sz/cz;  // Compute the intersection of the ray with the slab ...
    tz1 = (1 - sz)/cz;  // ... defined by the planes z = 0 and z = 1.
    float tLo = max(tMin, min(tz0,tz1));
    float tHi = min(tMax, max(tz0,tz1));
    if (tLo >= tHi || tHi <= 0)
       return false;  // Miss
    if (tLo > 0)
      inter.addHit(tLo, this);          // Add to intersection list
    inter.addHit(tHi, this);
    pLastRay = &r;                      // Save a pointer to the ray
    return true;
}

void CTaperedCylinder::getHitData(CRay& ray, float t, CVec3df& hitPoint, CVec3df& normal) {
   if (&ray != pLastRay)
      displayMessage(ERROR_MESSAGE, "Unexpected call to getHitData");
   hitPoint = ray.p(t);
   CVec3df genN(0,0,-1);  // Generic normal, assuming we hit the base
   if (t == tz1)          // Hit the end cap?
      genN = CVec3df(0,0,1);
   else if (t != tz0) {   // Hit the wall -- harder!
      CVec3df genP = transformedRay.p(t);
      genN = CVec3df(genP[X], genP[Y], (1 - s)*(1 - (1 - s)*genP[Z]));
   }
   normal = TInvTranspose.transformVector(genN).normalise();
}

// CSquare
// =======
bool CSquare::hit(CRay &ray, CIntersection &inter) {
   inter.clear();
   CRay transformedRay = ray.transformed(TInverse);
   // cout << "Transformed ray = " << transformedRay << endl;
     // Ray transformed back into primitive (unit square) space.
   float denom = transformedRay.dir[Z];
   if (fabs(denom) < 0.000001) return false;  // Parallel
   float t = -transformedRay.start[Z]/denom;
   CVec3df pHit = transformedRay.p(t);
   if (t <= 0 || fabs(pHit[X]) >= 0.5f || fabs(pHit[Y]) >= 0.5f)
        return false;
   inter.addHit(t, this);
   return true;
}

void CSquare::getHitData(CRay& ray, float t, CVec3df& hitPoint, CVec3df& hitNormal) {
   hitPoint =  ray.p(t);
   CVec3df genericNormal(0,0,1);
   hitNormal = TInvTranspose.transformVector(genericNormal).normalise();
}

// BSP Tree Stuff
// ==============

void CBSPTree::getFirstHit(CRay& ray, CIntersection& best, float tmin, float tmax) { // UDOO!!!
// Find the first object that the given ray hits within "this". Return
// a description of the hit in "best".
    best.clear();
    CVec3df n = plane.getNormal();
    float perpVelocity = ray.dir.dot(n);
    float startDistFromPlane = plane.distanceFromOrigin() - ray.start.dot(n);
    bool headingAway = perpVelocity > 0;
    bool startAbovePlane = startDistFromPlane < 0;
    bool startOnPlane = abs(startDistFromPlane) < 0.00001f;
    CBSPNode* nearTree = startAbovePlane ? outside : inside;
    CBSPNode* farTree = startAbovePlane ? inside : outside;
      
    float tWhenHitPlane = startDistFromPlane / perpVelocity;

    if (startOnPlane) {
        if (headingAway)
           outside->getFirstHit(ray, best, tmin, tmax);
        else
           inside->getFirstHit(ray, best, tmin, tmax);
    }
    else if (tWhenHitPlane > tmax || tWhenHitPlane < 0)
       nearTree->getFirstHit(ray, best, tmin, tmax);  // Whole subray is on near side
    else if (tWhenHitPlane < ray.tmin)
       farTree->getFirstHit(ray, best, tmin, tmax);   // Whole subray is on far side
    else {
       nearTree->getFirstHit(ray, best, tmin, tWhenHitPlane);
       if (best.numHits == 0) {
          farTree->getFirstHit(ray, best, tWhenHitPlane, tmax);
       }
    }
}


void CBSPLeaf::getFirstHit(CRay& ray, CIntersection& best, float tmin, float tmax){
// Finds the first object hit by ray within this leaf. 
// tmin and tmax specify the subrange of the ray that is of interest.
// [Actually, we disregard tmin, since it's there only for compatability
// with the internal node call, when it's needed for skipping the near plane
// calcs.]

     CIntersection inter;
     best.clear();        // No hits yet
     numLeafsVisited++;   // Gather stats
     int n = (int) objects->size();
     // cout << "Tracing leaf " << bbox << " numObjs = " << n << endl;
     for (int i = 0; i < n; i++) {
         if ((*objects)[i]->hit(ray, inter)) {    // If the ray hits this object
            if (best.numHits == 0 || inter.hit[0].t < best.hit[0].t) // And it's the best
               best.set(inter);
         }
     }
     if (best.numHits != 0 && best.hit[0].t > tmax)
          best.clear();     // Discard hits if out of bounds.
}
    

CBSPNode* makeBSPTree(ObjectList* objects, CBBox3f& box, int maxNumObjInLeaf, int maxDepth) {
// "Constructor" for a BSP tree, given a list of scene objects, the scene
// volume's bounding box, 
// a count of the desired maximum number of objects in each leaf (may
// not be achievable) and a desired maximum tree depth.   

// In this implementation, partitioning planes are axis aligned, perpendicular
// to the longest axis of the set of all scene objects. That's why the 
// bounding box parameter is required (or, at any rate, convenient!).

    int numObj = (int) objects->size(); 
    if ( numObj <= maxNumObjInLeaf || maxDepth <= 0)
        return new CBSPLeaf(objects, box);
        
    // Not the simple leaf case. So construct a partitioning plane,
    // partition the world, and build a BSP tree of the result.
    
    CPlane3df plane;
    CBBox3f newBox0(box), newBox1(box);
    int axis = box.longestAxis();
    if (axis == 0) {
           float xmid = (box.xmin + box.xmax)/2;
           plane.set(1,0,0,xmid);
           newBox0.xmax = xmid;
           newBox1.xmin = xmid;
    } else if (axis == 1) {
           float ymid = (box.ymin + box.ymax)/2;
           plane.set(0,1,0,ymid);
           newBox0.ymax = ymid;
           newBox1.ymin = ymid;
    } else { // axis = 2
           float zmid = (box.zmin + box.zmax)/2;
           plane.set(0,0,1,zmid);
           newBox0.zmax = zmid;
           newBox1.zmin = zmid;
    }

    ObjectList* insideObjects = new ObjectList;
    ObjectList* outsideObjects = new ObjectList;
    for (int i = 0; i< numObj; i++) { // Categorize all scene objects wrt plane
        CSceneObject* pObj = (*objects)[i];
        int whatSide = pObj->whatSideOf(plane);
        if (whatSide <= 0) insideObjects->push_back(pObj);
        if (whatSide >= 0) outsideObjects->push_back(pObj);
    }
    
    delete objects;
    
    //cout << "Partitioned with plane " << plane << ", numIn = "
    //    << (int) insideObjects->size() << ", numOut = " << (int) outsideObjects->size() << endl;

    CBSPNode* inTree = makeBSPTree(insideObjects, newBox0, maxNumObjInLeaf, maxDepth - 1);
    CBSPNode* outTree = makeBSPTree(outsideObjects, newBox1, maxNumObjInLeaf, maxDepth - 1);

    return new CBSPTree(plane, inTree, outTree, box);    
}    


// CScene
// ======

CScene::CScene(CSceneObject* firstObj, int numLights, 
                     CLight* lights, int maxNumObjInLeaf, int maxTreeDepth) {
    // Constructor for a scene.
    
    // First convert the linked list of objects to vector form,
    // constructing bounding box as we go.
    
    ObjectList* objects = new ObjectList;
    CSceneObject* pObj = firstObj;
	CBBox3f bbox;
    while (pObj != NULL) {
        objects->push_back(pObj);
		bbox.adjustToInclude(pObj->bbox);
        pObj = pObj->next;
    }
    
	cout << "Building BSP tree with " << (int) objects->size() << " objects\n";
    bsp = makeBSPTree(objects, bbox, maxNumObjInLeaf, maxTreeDepth);
	cout << "Done. Tree contains " << CBSPLeaf::numLeafs << " leaves, with an average of " 
 		<< (CBSPLeaf::numLeafObjects / (float) CBSPLeaf::numLeafs) << " objects per leaf.\n";
    this->numLights = numLights;
    this->lights = lights;
}

CColour CScene::shade(CRay& ray) {
// Computes the colour of the scene at a point as perceived by an eye
// looking along the given ray.
    CIntersection best, shadowInter;
    numShadeRays++;
    CColour clr(0,0,0);               // Colour of this ray.
    CColour background(0.9f,0.9f,0.9f);  // Background is a light grey
    getFirstHit(ray, best);
    if( best.numHits == 0) return background;  // Missed!
    CHitInfo& hit = best.hit[0];                // First intersection
    CMaterial& mat = hit.pObj->material;
    CVec3df hitPoint, hitNormal;  // point and normal at the ray-object hit point
    hit.pObj->getHitData(ray, hit.t, hitPoint, hitNormal);
    CVec3df rayStartPoint = hitPoint + hitNormal * 0.0001f;  // Offset shadow/refl ray start point
    for (int i = 0; i < numLights; i++) {
         clr += mat.ambient * lights[i].ambient;
         CVec3df lightDir = (lights[i].position - hitPoint).normalise();
         CRay shadowTestRay(rayStartPoint, lightDir);
         numShadowTestRays++;
         getFirstHit(shadowTestRay, shadowInter);
         if (shadowInter.numHits == 0) {
            CVec3df L = (lights[i].position - hitPoint).normalise();
            float sDotM = max(dot(L,hitNormal),0);
            CVec3df h = (L - ray.dir).normalise();
            float hDotM = max(dot(h, hitNormal),0);
            clr += mat.diffuse * lights[i].diffuse * sDotM;
            clr += mat.specular * lights[i].specular * (float) pow(hDotM, mat.shininess);
         }
    }

    if (mat.reflectivity > 0 && ray.recursionDepth < MAX_RECURSION_DEPTH) {
           CVec3df reflectedDir = ray.dir - 2* dot(ray.dir, hitNormal) * hitNormal;
           CRay reflectedRay(rayStartPoint, reflectedDir, ray.recursionDepth + 1);
           clr += mat.reflectivity * shade(reflectedRay);
    }
    return clr; 
}

void CScene::getFirstHit(CRay& ray, CIntersection& best) {
// Returns a description of the intersection between the given ray and the
// first object encountered in the scene.
    ray.clipToBBox(bsp->bbox);   // Clip ray to scene volume.
	ray.tmin -=0.00001f; ray.tmax += 0.00001f;  // Hack a bit to avoid cropping an axis aligned face
    // cout << "After clipping to bounding box, ray = " << ray << endl;
    if (!ray.isNull())
        bsp->getFirstHit(ray, best, ray.tmin, ray.tmax);
}
