// Header for the Scene Description classes including:
//
//        CMaterial
//        CLight
//        CScene
//        CSceneObject and its subclasses
//        BSP tree stuff

#if !defined(SCENE_DESCRIPTION_372FC)
#define SCENE_DESCRIPTION_372FC

#include "Geometry.h"
#include "Utilities.h"
#include "CColour.h"
#include "CTransformation.h"
#include "RayStuff.h"
#include <vector>

// The CMaterial class
// ===================
// A material is just a public set of reflectance properties. These are
// the same as in OpenGL except for the addition of:
// (a) "reflectivity", which is non-zero to add mirror reflection effects, and
// (b) "transparency", which is non-zero to add refracted ray effects.

class CMaterial {
  public:
     CColour ambient, diffuse, specular;
     float shininess;
     float reflectivity, transparency;
     CMaterial();
     CMaterial(CColour& ambient, CColour& diffuse, CColour& specular, float shininess);
     CMaterial(CColour& ambient, CColour& diffuse, CColour& specular,
          float shininess, float reflectivity, float transparency);
};


// The CLight class
// ================
// A Light is something that delivers illumination to a specified point
// on a surface. We could build a base class and specialize this for specific
// lights like PointLight, FloodLight, etc but I've chosen to implement CLight as
// just a PointLight. So it's really just a simple record, with all fields
// public.
class CLight {
  public:
     CVec3df position;
     CColour ambient;
     CColour diffuse;
     CColour specular;
	 CLight() {
		position = CVec3df(0,0,0);
		ambient = diffuse = specular = CColour(0,0,0);
	 }
     CLight(CVec3df& position, CColour& ambient, CColour& diffuse,
            CColour& specular) {
         this->position = position;
         this->ambient = ambient;
         this->diffuse = diffuse;
         this->specular = specular;
     }
};

// CSceneObject -- the generic Scene Object.
// ==========================================
// A generic CSceneObject is something that
// has a surface material and can be intersected with a Ray. It also has a
// transformation associated with it, which can be used to scale, translate,
// rotate, etc the scene object.
// It's an abstract class as a consequence of having a "pure virtual" method,
// i.e. a virtual method that is set to 0.
// [Don't blame me. I didn't design the language.]

// The biggest difference from Hill's approach is that I return minimal data
// on a call to hit -- if the hit turns out to be the closest, a follow up
// call to getHitData is made to get the details. The call to hit should try
// to optimise this process by caching any results that will be needed
// subsequently. This approach is more efficient than computing and saving
// results that won't ever be needed.

class CSceneObject {
  public:
     CMaterial material;
     CSceneObject* next;  // Next in list
     CBBox3f bbox;        // The object's bounding box
     CSceneObject() { next = NULL; };  // Subclass constructors call this one.
     static int numHitCalls;  // Statistics gathering
     void setTransformation(CTransformation& T);
     virtual bool hit(CRay &r, CIntersection &inter) = 0;
             // Returns true if the given ray hits the object, in which case
             // the given Intersection will contain the hit list for the object.
     virtual void getHitData(CRay& ray, float t, CVec3df& hitPoint, CVec3df& normal) = 0;
             // Called to get surface point and normal.
             // Valid only after a call to "hit" has returned true.
     virtual int whatSideOf(CPlane3df& plane);
             // Called to determine what side of the given plane this object lies.
             // Values are -1 for inside, +1 for outside, 0 for uncertain/straddle.
             // The default implementation uses the bounding box (which
             // the constructor should define).
  protected:
     CTransformation T;
     CTransformation TInverse; // Inverse of T
     CTransformation TInvTranspose; // Inverse transpose of T
     CRay transformedRay;     // The transformed ray in the last "hit" call.
     CRay* pLastRay;          // Sanity check (see getHitData).
};

typedef vector<CSceneObject*> ObjectList;  // Define the "list of scene objects" type

// The CSphere class, a concrete instance of CSceneObject
// ==================
class CSphere : public CSceneObject {
  public:
     CSphere(CVec3df& centre, float radius, CMaterial& mat);
     CSphere(CTransformation& M, CMaterial& mat);
     virtual bool hit(CRay &r, CIntersection &inter);
     virtual void getHitData(CRay& ray, float t, CVec3df& hitPoint, CVec3df& normal);
             // Called to determine what side of the given plane this object lies.
             // Values are -1 for inside, +1 for outside, 0 for uncertain/straddle
};

// The CTaperedCylinder class.
// ==========================
// This shape is as described by Hill -- a tapered cylinder along the z axis
// with a unit radius base at z = 0 and a top of radius "s" at z = 1.
class CTaperedCylinder : public CSceneObject {
  public:
     CTaperedCylinder(float s, CTransformation& M, CMaterial& mat);
     virtual bool hit(CRay &r, CIntersection &inter);
     virtual void getHitData(CRay& ray, float t, CVec3df& hitPoint, CVec3df& normal);
             // Called to determine what side of the given plane this object lies.
             // Values are -1 for inside, +1 for outside, 0 for uncertain/straddle
  private:
     float s; // Radius of the top cap (following Hill's notation)
     float tz0, tz1;  // Hit distances to base and top planes (cached for getHitInfo)
};

// The CSquare class
// ====================
// A unit square in the z = 0 plane, centred on the origin.
// You can transform it into something slightly more interesting but not very.
class CSquare : public CSceneObject {
  public:
     CSquare(CTransformation tx, CMaterial& mat) {
		this->material = mat;
	    bbox.set(-0.5f,0.5f,-0.5f,0.5f,-0.00001f,0.00001f);
		setTransformation(tx);
	 }
     bool hit(CRay& ray, CIntersection& inter);
     void getHitData(CRay& ray, float t, CVec3df& hitPoint, CVec3df& hitNormal);
};

// The CBSPNode class.
// ==================
// This is the abstract superclass for CBSPLeaf and CBSPTree.
class CBSPNode {
  public:
     CBBox3f bbox;
     virtual void getFirstHit(CRay& ray, CIntersection& best, float tmin, float tmax) = 0;
};

// The CBSPTree class, which is the internal node of a BSPTree.
// ==================
class CBSPTree : public CBSPNode {
   public:
      CPlane3df plane;      // The partitioning plane
      CBSPNode* inside;     // The description of the world inside the plane
      CBSPNode* outside;    // The description of the world outside the plane
      CBSPTree(CPlane3df& plane, CBSPNode* in, CBSPNode* out, CBBox3f box) {
            this->plane = plane;
            inside = in;
            outside = out;
            bbox = box;
      }
      void getFirstHit(CRay& ray, CIntersection& best, float tmin, float tmax);
};

// The CBSPLeaf class, representing a leaf of a BSP tree.
// ==================
// A leaf is just a list of scene objects.
class CBSPLeaf : public CBSPNode {
   public:
      ObjectList* objects;
	  static int numLeafs;	// No. of leaf nodes created
	  static int numLeafObjects;  // No. of objects (for average calculation)
      CBSPLeaf(ObjectList* objs, CBBox3f box) {
            objects = objs;
            // cout << "Leaf created with " << (int) objects->size() << " objects\n";
            bbox = box;
		    numLeafs++;
			numLeafObjects += (int) objects->size();
      }
      static int numLeafsVisited;  // Statistics gathering -- counts "getFirstHit" calls.
      void getFirstHit(CRay& ray, CIntersection& best, float tmin, float tmax);
}; 

// "Constructor" for a BSP tree, given a list of scene objects, the scene
// volume's bounding box, a count of the desired maximum number of objects in each leaf (may
// not be achievable) and a desired maximum tree depth.    
CBSPNode* makeBSPTree(ObjectList* objects, CBBox3f& box, int maxNumObjInLeaf, int maxDepth);
  

// CScene class
// ============

// A Scene is a scene graph plus lighting and rendering information.
// In this version it builds a BSP tree of the scene volume.

class CScene {
  public:
     CScene() { bsp = NULL; numLights = 0; }
     CScene(CSceneObject* scene, int numLights, CLight* lights, int maxNumObjInLeaf, int maxTreeDepth);
     CColour shade(CRay& ray);  // The colour of the scene, looking in along "ray".
     static int numShadeRays;
     static int numShadowTestRays;
  private:
    void getFirstHit(CRay& ray, CIntersection& best);
    // Returns in "best" the description of the first intersection between the
    // given ray and the scene.
    CBSPNode* bsp;    // The BSP tree containing all the scene objects
    int numLights;    // The number of lights in lights[]
    CLight* lights;   // The array of lights.
};

#endif

